import greenfoot.*;
import java.util.HashMap;

public class Player extends Actor{
    private static final ratio[] speed = new ratio[]{
        new ratio(1,2),new ratio(1),new ratio(3,2)}; //Must be positive : infinite loop
    private static final ratio[] jumpheight = new ratio[]{
        new ratio(-6),new ratio(-7),new ratio(-8),new ratio(-9),new ratio(-10)};
    
    public Image[] HP = new Image[3];
    public Player(String name){
        setupImages(name);
        this.name = name;
        for(int i = 0; i < HP.length; i++) HP[i] = new Image("heart.png",30,30);
    }
    
    private String name = "";
    @Override public String toString(){ return name; }
    
// MAIN
    public void act(){
        Input();
        Touch();
    }
    
    
    
// PHYSICS RELATED
    private MouseInfo mouse = null;
    private void Input(){
        mouse = Greenfoot.getMouseInfo();
        if(Greenfoot.mousePressed(null)) mousePressed = true;
        else if(Greenfoot.mouseClicked(null)) mousePressed = false;
        byte dx = 0;
        byte dy = 0;
        if(mouse != null && mousePressed){
            int x = getX() - mouse.getX();
            int y = getY() - mouse.getY();
            double h = Math.sqrt(Math.pow(x,2)+Math.pow(y,2));
            if(25 < h){
                double angle = Math.acos((double)x/h);
                if(y < 0) angle = 2* Math.PI - angle;
                angle = (angle + Math.PI/8) % (2*Math.PI); //round
                int sector = (int)(4*angle/Math.PI);
                switch(sector){
                    case 0: dx =-1; dy = 0; break;
                    case 1: dx =-1; dy = 1; break;
                    case 2: dx = 0; dy = 1; break;
                    case 3: dx = 1; dy = 1; break;
                    case 4: dx = 1; dy = 0; break;
                    case 5: dx = 1; dy =-1; break;
                    case 6: dx = 0; dy =-1; break;
                    case 7: dx =-1; dy =-1; break;
                    default:dx = 0; dy = 0; break;
                }
            } dx += dx*Math.min(speed.length-1,h/50); 
              dy += dy*Math.min(jumpheight.length-1,h/25);
        } Logic(dx,dy);
        Anime(dx,dy);
    }
    private ratio yVelocity = new ratio(0);
    private void Logic(int inputx,int inputy){
        int x = 0; 
        int y = 0;
        int dx = inputx; //rename variable
        inputx = Math.abs(inputx);
        for(x = 0; dx != 0; x += dx){
            if(Math.abs(speed[inputx-1].doubleValue()*dx) < Math.abs(x)) dx = 0;
            else for(int p = -1; p <= 1; p++){
                if(getOneObjectAtOffset(10*dx,p,Collidable.class)!=null)
                    dx = 0;
            }
        } yVelocity.add(Zone.GRAVITY);
        boolean onFloor = false;
        int dy = ratio.signum(yVelocity);
        for(y = 0; dy != 0; y += dy){
            if(Math.abs(yVelocity.intValue())< Math.abs(y)) dy = 0;
            else for(int p = -1; p <= 1; p++){
                int radius = getImage().getWidth()/2;
                if(getOneObjectAtOffset((radius-1)*p-p,radius*dy+y,Collidable.class)!=null){
                    if(0 < dy) onFloor = true;
                    dy = 0;
                }
            }
        } if(onFloor){
            if(0 < inputy) yVelocity.set(jumpheight[Math.abs(inputy)-1]); 
            else yVelocity = new ratio(0);
        } ((Zone)getWorld()).move(x,y);
    }
    
// COLLISION
    private int[] DMGDELAY = {0,64};
    private void Touch(){
        boolean damage = false;
        if(0 < DMGDELAY[0]) DMGDELAY[0]--;
        else if(getOneObjectAtOffset(0,getImage().getHeight()/2,Spike.class) != null){
            for(int i = HP.length-1; i >= 0 && !damage; i--)
                if(HP[i] != null){
                    getWorld().removeObject(HP[i]);
                    HP[i] = null;
                    damage = true;
                    yVelocity.set(ratio.pro(jumpheight[4],new ratio(3,2)));
                    if(i == 0){ // respawn character at 0 HP
                        for(int j = 0; j < HP.length; j++) HP[j] = new Image("heart.png",30,30);
                        Zone world = (Zone)getWorld();
                        world.load(world.autosave[0],world.autosave[1]);
                    }
                }
            }
        if(damage) DMGDELAY[0] = DMGDELAY[1];
    }
    
    
    
// IMAGE METHODS
    private int anime = 0;
    private int face = 1;
    private GreenfootImage[] idleL = new GreenfootImage[4];
    private GreenfootImage[] idleR = new GreenfootImage[4];
    private GreenfootImage[] runL = new GreenfootImage[6];
    private GreenfootImage[] runR = new GreenfootImage[6];
    public void Anime(int dx,int dy){
        int ANIMEDELAY = 3;
        int direction = Integer.signum(dx);
        if(direction == -face) face = direction;
        else if(dx != 0){
            anime = (anime+1) % (ANIMEDELAY*runR.length);
            if(face == 1) setImage(runR[anime/ANIMEDELAY]);
            else if(face == -1) setImage(runL[anime/ANIMEDELAY]);
        } else{
            anime = (anime+1) % (ANIMEDELAY*idleR.length);
            if(face == 1) setImage(idleR[anime/ANIMEDELAY]);
            else if(face == -1) setImage(idleL[anime/ANIMEDELAY]);
        }
    }
    private void setupImages(String filename){
        GreenfootImage slide = new GreenfootImage("sprites"+File.IN+"sheets"+File.IN+filename+".png");
          for(int i = 0; i < idleR.length; i++){
            idleR[i] = new GreenfootImage(24,24);
            idleL[i] = new GreenfootImage(24,24);
            for(int y = 0; y < 24; y++) for(int x = 0; x < 24; x++){
                idleR[i].setColorAt(x,y,slide.getColorAt(24*i+x,y));
                idleL[i].setColorAt(23-x,y,slide.getColorAt(24*i+x,y));
            } idleR[i].scale(Zone.UNITSIZE,Zone.UNITSIZE);
              idleL[i].scale(Zone.UNITSIZE,Zone.UNITSIZE);
        } for(int i = 0; i < runR.length; i++){
            runR[i] = new GreenfootImage(24,24);
            runL[i] = new GreenfootImage(24,24);
            for(int y = 0; y < 24; y++) for(int x = 0; x < 24; x++){
                runR[i].setColorAt(x,y,slide.getColorAt(24*(i+4)+x,y));
                runL[i].setColorAt(23-x,y,slide.getColorAt(24*(i+4)+x,y));
            } runR[i].scale(Zone.UNITSIZE,Zone.UNITSIZE);
              runL[i].scale(Zone.UNITSIZE,Zone.UNITSIZE);
        }
    }
    
    
    
// USER INTERFACE
    private HashMap<String,String> keys = new HashMap<String,String>();
    private boolean mousePressed = false;
    public void configure(String command,String key){
        if(!keys.keySet().contains(command)) keys.put(command,key);
        else keys.replace(command,key);
    }
}
