/**
  * class ratio:
  *   The ratio class constructs fractions, which include an integer numerator and denominator.
  *   This class includes basic mathematical operations, as well as some conversions between decimals and fractions.
 **/
public class ratio extends Number{
    //MAIN MEMBERS
  public int numer;
  public int denom;
  
  
    //CONSTRUCTORS
  public ratio(){ this(0); }
  public ratio(int numer){ this.numer = numer; this.denom = 1; }
  public ratio(int numer, int denom){ this.numer = numer; this.denom = denom; }
  public ratio(double decimal){ this(toRatio(decimal)); }
  public ratio(double decimal,int denom){ this(toRatio(decimal,denom)); }
  public ratio(String ratio_S){ this(toRatio(ratio_S)); }
  public ratio(ratio that){ this.numer = that.numer; this.denom = that.denom; }
  
  
    //RECONSTRUCTORS
  public void set(int numer,int denom){ this.numer = numer; this.denom = denom; }
  public void set(ratio that){ this.numer = that.numer; this.denom = that.denom; }
  public ratio copy(ratio that){ return new ratio(that); }
  
  
    //OBJECT INHERITED METHODS
  @Override public String toString(){ 
      return (denom == 0)? (numer == 0)? "undefined" : 
             (((0 < Integer.signum(numer))?"+":"-")+"infinite") :
             (numer == 0 || denom == 1)? numer + "" :
             (numer % denom == 0)? (numer / denom) + "" :numer + "/" + denom;
  }
  @Override public boolean equals(Object that){
      return equals(toRatio(that.toString()));
  }
  
  
    //MATH: COMPARISONS
  public boolean equals(ratio that){
    ratio _that = new ratio(that);
    ratio _this = new ratio(this);
    _this = reduce(_this); _that = reduce(_that);
    return _this.numer == _that.numer && _this.denom == _that.denom;
  }
  public boolean isLess(ratio that){
    if(this.denom == that.denom) return this.numer < that.numer;
    else{
      long this_numer = (long)this.numer * that.denom;
      long that_numer = (long)that.numer * this.denom;
      return this_numer < that_numer;
    }
  }
  public boolean isMore(ratio that){
    if(this.denom == that.denom) return this.numer > that.numer;
    else{
      long this_numer = (long)this.numer * that.denom;
      long that_numer = (long)that.numer * this.denom;
      return this_numer > that_numer;
    }
  }
  public static ratio min(ratio a,ratio b){ return (a.isLess(b))? a : (b.isLess(a))? b : null; }
  public static ratio max(ratio a,ratio b){ return (a.isMore(b))? a : (b.isMore(a))? b : null; }
  public static ratio abs(ratio fraction){return new ratio(Math.abs(fraction.numer),Math.abs(fraction.denom));}
  
  
    //MATH: INCREMENTATIVE OPERATIONS
  public ratio add(ratio that){ ratio sum = ratio.sum(this,that); set(sum); return sum; }
  public ratio add(int numer,int denom){ ratio sum = ratio.sum(this,new ratio(numer,denom)); set(sum); return sum; }
  public ratio sub(ratio that){ return add(that.opposite()); }
  public static ratio sum(ratio a,ratio b){
    if(a.denom == b.denom)
      return new ratio(a.numer + b.numer,a.denom);
    else{
      long numer  = ((long)a.numer * (long)b.denom) + ((long)b.numer * (long)a.denom);
      long denom  =  (long)a.denom * (long)b.denom;
      int factor = 1;
      long max = Math.max(numer, denom);
      for (int f = 1; f <= max; f++)
        if ((numer % f == 0) && (denom % f == 0))
          factor = f;
      numer /= factor; denom /= factor;
      return new ratio((int)numer,(int)denom);
    }
  }
  public ratio mul(ratio that){ ratio pro = ratio.pro(this,that); set(pro); return pro; }
  public ratio mul(int numer,int denom){ ratio pro = ratio.pro(this,new ratio(numer,denom)); set(pro); return pro; }
  public ratio div(ratio that){ return mul(copy(that).flip()); }
  public static ratio pro(ratio a,ratio b){
    int sign = Integer.signum(a.numer)*Integer.signum(a.denom)*Integer.signum(b.numer)*Integer.signum(b.denom);
    long numer = Math.abs(a.numer * b.numer); long denom = Math.abs(a.denom * b.denom);
    int factor = 1;
    long max = Math.max(numer, denom);
    for (int f = 1; f <= max; f++)
      if ((numer % f == 0) && (denom % f == 0))
        factor = f;
    numer /= factor; denom /= factor;
    return new ratio(sign * (int)numer,(int)denom);
  }
  public ratio pow(int expo){
    if(expo == 0) return new ratio();
    else if(expo < 0) flip();
    expo = Math.abs(expo);
    if(expo == 1) return this;
    reduce();
    java.math.BigInteger numerB = new java.math.BigInteger(this.numer + "");
    java.math.BigInteger denomB = new java.math.BigInteger(this.denom + "");
    long numer = numerB.pow(expo).longValue(); long denom = denomB.pow(expo).longValue();
    int factor = 1;
    long max = Math.max(numer, denom);
    for(int f = 1; f <= max; f++)
      if((numer % f == 0) && (denom % f == 0))
        factor = f;
    numer /= factor; denom /= factor;
    return new ratio((int)numer,(int)denom);
  }
  
  
    //RATIO: FRACTION OPERATIONS
  public static int signum(ratio R){ 
      ratio zero = new ratio(0); 
      return (R.isLess(zero))? -1 : (R.isMore(zero))? 1 : 0;
  }
  public ratio opposite(){ return opposite(this); }
  public ratio opposite(ratio R){ ratio r = copy(R); r.numer *= -1; return r; }
  public ratio flip(){ return flip(this); }
  public static ratio flip(ratio R){ int _numer = R.numer; byte sign = (byte)Integer.signum(_numer); R.numer = sign * R.denom; R.denom = sign * _numer; return R;}
  public ratio reduce(){ return reduce(this); }
  public static ratio reduce(ratio R) {
    if (R.numer == 0) {
      R = new ratio(0);
    } else {
      int sign = (Integer.signum(R.numer) != Integer.signum(R.denom))? -1 : 1;
      int numer = Math.abs(R.numer); int denom = Math.abs(R.denom);
      int factor = 1;
      int max = Math.max(numer, denom);
      for (int f = 1; f <= max; f++)
        if ((numer % f == 0) && (denom % f == 0))
          factor = f;
      numer /= factor; denom /= factor;
      R.set(sign * numer, denom);
    } return R;
  }
  
  
    //Number methods
  public int intValue(){ return numer / denom; }
  public static int intValue(ratio fraction){ return fraction.numer / fraction.denom; }
  public long longValue(){ return (long)intValue(); }
  public static long longValue(ratio fraction){ return (long)intValue(fraction); }
  public float floatValue(){ return (float)doubleValue(); }
  public static float floatValue(ratio fraction){ return (float)doubleValue(fraction); }
  public double doubleValue(){ return toDouble(this,precision); }
  public static double doubleValue(ratio fraction){ return toDouble(fraction,precision); }

  
    //Double methods
  private static byte precision = 2;
  public static void setPrecision(int precision){
    if(0 <= precision && precision <= 8) ratio.precision = (byte)precision;
  }
  public static byte getPrecision(){
    return precision;
  }
  public static ratio toRatio(double decimal){ return toRatio(decimal,(int)Math.pow(10,precision)); }
  public static ratio toRatio(double decimal,int denom){
    int sign = (int)java.lang.Math.signum(decimal); 
    decimal *= sign;
    int numer = (int)decimal * denom;
    decimal -= (int)decimal;
    int numerMinError = 0; double valueMinError = 1;
    for(int i = 0; i < denom; i++){
      double valueError = Math.abs(((double)i/denom) - decimal);
      if(valueError < valueMinError){
        numerMinError = i; valueMinError = valueError;
      }
    } numer += numerMinError;
    return new ratio(sign * numer,denom);
  }
  public double toDouble(int precision){ return toDouble(this,precision); }
  public static double toDouble(ratio fraction,int precision){
    int sign = Integer.signum((int)fraction.numer)*Integer.signum((int)fraction.denom);
    int whole = sign * fraction.numer / fraction.denom;
    int numer = Math.abs(fraction.numer) % Math.abs(fraction.denom); int denom = Math.abs(fraction.denom);
    String decimal = "";
    for(byte i = 1; i <= precision; i++){
      numer *= 10;
      byte digit = 0;
      for(byte iDigit = 0; iDigit <= 9; iDigit++)
        if((denom * iDigit) <= numer)
          digit = iDigit;
      decimal += digit;
      numer -= (denom * digit);
    } decimal = whole + "." + decimal;
    return sign * Double.parseDouble(decimal);
  }
  
    //String methods
  public static ratio toRatio(String ratio_S){
    ratio_S = ratio_S.replaceAll("\\s+","");
    String[] ratio_split = new String[2];
    if(ratio_S.contains("/"))
      ratio_split = ratio_S.split("/");
    else{ ratio_split[0] = ratio_S; ratio_split[1] = "1"; }
    ratio R = null;
    try{ R = new ratio(Integer.parseInt(ratio_split[0]),Integer.parseInt(ratio_split[1])); }
    catch(Exception e){ System.err.println("could not convert String to ratio: " + ratio_S); }
    return R;
  }
  public static ratio[] parseRatios(String ratio_array){
    ratio_array = ratio_array.replaceAll("\\s","");
    String[] ratios_S = ratio_array.split(",");
    ratio[] ratios = new ratio[ratios_S.length];
    for(short i = 0; i < ratios_S.length; i++)
      ratios[i] = toRatio(ratios_S[i]);
    return ratios;
  }
}